"""This module defines the ChEMBL ETL methods."""
from .base import Base
from therapy import PROJECT_ROOT
from therapy.schemas import SourceName, NamespacePrefix, ApprovalStatus, \
    SourceMeta
from typing import List
from ftplib import FTP
import logging
import tarfile
import sqlite3
import os
import shutil

logger = logging.getLogger('therapy')
logger.setLevel(logging.DEBUG)


class ChEMBL(Base):
    """ETL the ChEMBL source into therapy.db."""

    def __init__(self,
                 database,
                 data_path=PROJECT_ROOT / 'data' / 'chembl' / 'chembl_27.db'):
        """Initialize CHEMBl ETL instance.

        :param Path data_path: path to CHEMBl source SQLite3 database file.
        """
        self.database = database
        self._data_path = data_path
        self._added_ids = []

    def perform_etl(self) -> List[str]:
        """Public-facing method to initiate ETL procedures on given data.

        :return: List of concept IDs which were successfully processed and
            uploaded.
        """
        self._load_meta()
        self._extract_data()
        self._transform_data()
        self._load_json()
        self._conn.commit()
        self._conn.close()
        return self._added_ids

    def _extract_data(self, *args, **kwargs):
        """Extract data from the ChEMBL source."""
        logger.info('Extracting chembl_27.db...')
        if 'data_path' in kwargs:
            chembl_db = kwargs['data_path']
        else:
            chembl_db = PROJECT_ROOT / 'data' / 'chembl' / 'chembl_27.db'
        if not chembl_db.exists():
            chembl_dir = PROJECT_ROOT / 'data' / 'chembl'
            chembl_archive = PROJECT_ROOT / 'data' / 'chembl' / 'chembl_27_' \
                                                                'sqlite.tar.gz'
            chembl_archive.parent.mkdir(exist_ok=True, parents=True)
            self._download_chembl_27(chembl_archive)
            tar = tarfile.open(chembl_archive)
            tar.extractall(path=PROJECT_ROOT / 'data' / 'chembl')
            tar.close()

            # Remove unused directories and files
            chembl_27_dir = chembl_dir / 'chembl_27'
            temp_chembl = chembl_27_dir / 'chembl_27_sqlite' / 'chembl_27.db'
            chembl_db = chembl_dir / 'chembl_27.db'
            shutil.move(temp_chembl, chembl_db)
            os.remove(chembl_archive)
            shutil.rmtree(chembl_27_dir)
        conn = sqlite3.connect(chembl_db)
        conn.row_factory = sqlite3.Row
        self._conn = conn
        self._cursor = conn.cursor()
        assert chembl_db.exists()
        logger.info('Finished extracting chembl_27.db.')

    def _transform_data(self, *args, **kwargs):
        """Transform SQLite data to JSON."""
        self._create_dictionary_synonyms_table()
        self._create_trade_names_table()
        self._create_temp_table()

        self._cursor.execute("DROP TABLE DictionarySynonyms;")
        self._cursor.execute("DROP TABLE TradeNames;")

    @staticmethod
    def _download_chembl_27(filepath):
        """Download the ChEMBL source."""
        logger.info('Downloading ChEMBL v27, this will take a few minutes.')
        try:
            with FTP('ftp.ebi.ac.uk') as ftp:
                ftp.login()
                logger.debug('FTP login completed.')
                ftp.cwd('pub/databases/chembl/ChEMBLdb/releases/chembl_27')
                with open(filepath, 'wb') as fp:
                    ftp.retrbinary('RETR chembl_27_sqlite.tar.gz', fp.write)
            logger.info('Downloaded ChEMBL tar file.')
        except TimeoutError:
            logger.error('Connection to EBI FTP server timed out.')

    def _create_dictionary_synonyms_table(self):
        """Create temporary table to store drugs and their synonyms."""
        create_dictionary_synonyms_table = f"""
            CREATE TEMPORARY TABLE DictionarySynonyms AS
            SELECT
                md.molregno,
                '{NamespacePrefix.CHEMBL.value}:'||md.chembl_id as chembl_id,
                md.pref_name,
                md.max_phase,
                md.withdrawn_flag,
                group_concat(
                    ms.synonyms, '||')as synonyms
            FROM molecule_dictionary md
            LEFT JOIN molecule_synonyms ms USING(molregno)
            GROUP BY md.molregno
            UNION ALL
            SELECT
                md.molregno,
                '{NamespacePrefix.CHEMBL.value}:'||md.chembl_id as chembl_id,
                md.pref_name,
                md.max_phase,
                md.withdrawn_flag,
                group_concat(
                    ms.synonyms, '||') as synonyms
            FROM molecule_synonyms ms
            LEFT JOIN molecule_dictionary md USING(molregno)
            WHERE md.molregno IS NULL
            GROUP BY md.molregno
        """
        self._cursor.execute(create_dictionary_synonyms_table)

    def _create_trade_names_table(self):
        """Create temporary table to store trade name data."""
        create_trade_names_table = """
            CREATE TEMPORARY TABLE TradeNames AS
            SELECT
                f.molregno,
                f.product_id,
                group_concat(
                    p.trade_name, '||') as trade_names
            FROM formulations f
            LEFT JOIN products p
                ON f.product_id = p.product_id
            GROUP BY f.molregno;
        """
        self._cursor.execute(create_trade_names_table)

    def _create_temp_table(self):
        """Create temporary table to store therapies data."""
        create_temp_table = """
            CREATE TEMPORARY TABLE temp(concept_id, label, approval_status,
                                        src_name, trade_names, aliases);
        """
        self._cursor.execute(create_temp_table)

        insert_temp = f"""
            INSERT INTO temp(concept_id, label, approval_status, src_name,
                             trade_names, aliases)
            SELECT
                ds.chembl_id,
                ds.pref_name,
                CASE
                    WHEN ds.withdrawn_flag
                        THEN '{ApprovalStatus.WITHDRAWN.value}'
                    WHEN ds.max_phase == 4
                        THEN '{ApprovalStatus.APPROVED.value}'
                    WHEN ds.max_phase == 0
                        THEN NULL
                    ELSE
                        '{ApprovalStatus.INVESTIGATIONAL.value}'
                END,
                '{SourceName.CHEMBL.value}',
                t.trade_names,
                ds.synonyms
            FROM DictionarySynonyms ds
            LEFT JOIN TradeNames t USING(molregno)
            GROUP BY ds.molregno
            UNION ALL
            SELECT
                ds.chembl_id,
                ds.pref_name,
                CASE
                    WHEN ds.withdrawn_flag
                        THEN '{ApprovalStatus.WITHDRAWN.value}'
                    WHEN ds.max_phase == 4
                        THEN '{ApprovalStatus.APPROVED.value}'
                    WHEN ds.max_phase == 0
                        THEN NULL
                    ELSE
                        '{ApprovalStatus.INVESTIGATIONAL.value}'
                END,
                '{SourceName.CHEMBL.value}',
                t.trade_names,
                ds.synonyms
            FROM TradeNames t
            LEFT JOIN DictionarySynonyms ds USING(molregno)
            WHERE ds.molregno IS NULL
            GROUP BY ds.molregno;
        """
        self._cursor.execute(insert_temp)

    def _load_json(self):
        """Load ChEMBL data into database."""
        chembl_data = """
            SELECT
                concept_id,
                label,
                approval_status,
                src_name,
                trade_names,
                aliases
            FROM temp;
        """
        result = [dict(row) for row in
                  self._cursor.execute(chembl_data).fetchall()]
        self._cursor.execute("DROP TABLE temp;")

        with self.database.therapies.batch_writer() as batch:
            for record in result:
                if record['label']:
                    self._load_label(record, batch)
                else:
                    del record['label']
                if record['aliases']:
                    self._load_alias(record, batch)
                else:
                    del record['aliases']
                if record['trade_names']:
                    self._load_trade_name(record, batch)
                else:
                    del record['trade_names']
                if not record['approval_status']:
                    del record['approval_status']
                self._load_therapy(record, batch)

    def _load_therapy(self, record, batch):
        """Load therapy record into DynamoDB."""
        record['label_and_type'] = \
            f"{record['concept_id'].lower()}##identity"
        record['item_type'] = 'identity'
        batch.put_item(Item=record)
        self._added_ids.append(record['concept_id'])

    def _load_label(self, record, batch):
        """Load label record into DynamoDB."""
        label = {
            'label_and_type':
                f"{record['label'].lower()}##label",
            'concept_id': f"{record['concept_id'].lower()}",
            'src_name': SourceName.CHEMBL.value,
            'item_type': 'label',
        }
        batch.put_item(Item=label)

    def _load_alias(self, record, batch):
        """Load alias records into DynamoDB."""
        record['aliases'] = record['aliases'].split("||")

        # Remove duplicates (case-insensitive)
        aliases = set({t.casefold(): t for t in record['aliases']})

        if len(aliases) > 20:
            del record['aliases']
        else:
            for alias in aliases:
                alias = {
                    'label_and_type': f"{alias}##alias",
                    'concept_id': f"{record['concept_id'].lower()}",
                    'src_name': SourceName.CHEMBL.value,
                    'item_type': 'alias',
                }
                batch.put_item(Item=alias)

            record['aliases'] = list(set(record['aliases']))

    def _load_trade_name(self, record, batch):
        """Load trade name records into DynamoDB."""
        record['trade_names'] = \
            record['trade_names'].split("||")

        # Remove duplicates (case-insensitive)
        trade_names = \
            set({t.casefold(): t for t in record['trade_names']})

        if len(trade_names) > 20:
            del record['trade_names']
        else:
            for trade_name in trade_names:
                trade_name = {
                    'label_and_type':
                        f"{trade_name}##trade_name",
                    'concept_id': f"{record['concept_id'].lower()}",
                    'src_name': SourceName.CHEMBL.value,
                    'item_type': 'trade_name'
                }
                batch.put_item(Item=trade_name)

            record['trade_names'] = list(set(record['trade_names']))

    def _load_meta(self, *args, **kwargs):
        """Add ChEMBL metadata."""
        metadata = SourceMeta(data_license='CC BY-SA 3.0',
                              data_license_url='https://creativecommons.org/licenses/by-sa/3.0/',  # noqa: E501
                              version='27',
                              data_url='http://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_27/',  # noqa: E501
                              rdp_url='http://reusabledata.org/chembl.html',
                              data_license_attributes={
                                  'non_commercial': False,
                                  'share_alike': True,
                                  'attribution': True
                              })
        params = dict(metadata)
        params['src_name'] = SourceName.CHEMBL.value
        self.database.metadata.put_item(Item=params)



"""A base class for extraction, transformation, and loading of data."""
from abc import ABC, abstractmethod
from typing import List, Dict
from therapy import ACCEPTED_SOURCES, PROJECT_ROOT
from therapy.schemas import Drug
import logging


logger = logging.getLogger('therapy')
logger.setLevel(logging.DEBUG)


class Base(ABC):
    """The ETL base class."""

    def __init__(self, database, data_path=PROJECT_ROOT / 'data'):
        """Extract from sources.

        :param Database database: application database object
        :param Path data_path: path to normalizer data directory
        """
        name = self.__class__.__name__.lower()
        self.database = database
        self._src_data_dir = data_path / name
        self._in_normalize = name in ACCEPTED_SOURCES
        if self._in_normalize:
            self._added_ids = []

    @abstractmethod
    def perform_etl(self) -> List[str]:
        """Public-facing method to begin ETL procedures on given data.

        Returned concept IDs can be passed to Merge method for computing
        merged concepts.

        :return: list of concept IDs which were successfully processed and
            uploaded.
        """
        raise NotImplementedError

    def _download_data(self, *args, **kwargs):
        raise NotImplementedError

    def _extract_data(self):
        """Get source file from data directory."""
        self._src_data_dir.mkdir(exist_ok=True, parents=True)
        src_file_prefix = f'{type(self).__name__.lower()}_'
        dir_files = [f for f in self._src_data_dir.iterdir()
                     if f.name.startswith(src_file_prefix)]
        if len(dir_files) == 0:
            self._download_data()
            dir_files = [f for f in self._src_data_dir.iterdir()
                         if f.name.startswith(src_file_prefix)]
        self._src_file = sorted(dir_files, reverse=True)[0]
        self._version = self._src_file.stem.split('_', 1)[1]

    @abstractmethod
    def _transform_data(self, *args, **kwargs):
        raise NotImplementedError

    def _load_therapy(self, therapy: Dict):
        """Load individual therapy record.

        :param Dict therapy: valid therapy object.
        """
        assert Drug(**therapy)
        concept_id = therapy['concept_id']

        field_pairs = (('aliases', 'alias'), ('xrefs', 'xref'),
                       ('associated_with', 'associated_with'),
                       ('trade_names', 'trade_name'))
        for field, field_name in field_pairs:
            if field in therapy:
                items = therapy[field]
                if items == [] or items is None:
                    del therapy[field]
                else:
                    items = {i.casefold() for i in items}
                    if field in ('aliases', 'trade_names'):
                        if len(items) > 20:
                            logger.debug(f"{concept_id} has > 20 {field}.")
                            del therapy[field]
                            continue
                    for item in items:
                        self.database.add_ref_record(item, concept_id,
                                                     field_name)

        if 'approval_status' in therapy \
                and therapy['approval_status'] is None:
            del therapy['approval_status']
        if 'label' in therapy:
            label = therapy['label']
            if label:
                self.database.add_ref_record(label, concept_id, 'label')
            else:
                del therapy['label']

        self.database.add_record(therapy)
        if self._in_normalize:
            self._added_ids.append(concept_id)

    @abstractmethod
    def _load_meta(self, *args, **kwargs):
        raise NotImplementedError



"""Provide ETL methods for HemOnc.org data."""
from therapy import DownloadException, PROJECT_ROOT
from therapy.etl.base import Base
from therapy.schemas import NamespacePrefix, SourceMeta, SourceName, \
    ApprovalStatus
from disease.query import QueryHandler as DiseaseNormalizer
from pathlib import Path
from typing import List, Dict, Tuple
import csv
import logging


logger = logging.getLogger('therapy')
logger.setLevel(logging.DEBUG)


class HemOnc(Base):
    """Docstring"""

    def __init__(self, database, data_path: Path = PROJECT_ROOT / 'data'):
        """Initialize HemOnc instance.

        :param therapy.database.Database database: application database
        :param Path data_path: path to normalizer data directory
        """
        super().__init__(database, data_path)
        self.disease_normalizer = DiseaseNormalizer(self.database.endpoint_url)

    def perform_etl(self) -> List[str]:
        """Public-facing method to begin ETL procedures on given data.

        Returned concept IDs can be passed to Merge method for computing
        merged concepts.

        :return: list of concept IDs which were successfully processed and
            uploaded.
        """
        self._extract_data()
        self._load_meta()
        self._transform_data()
        return self._added_ids

    def _download_data(self):
        """Download HemOnc.org source data.

        Raises download exception for now -- HTTP authorization may be
        possible?
        """
        raise DownloadException("No download for HemOnc data available -- "
                                "must place manually in data/ directory.")

    def _extract_data(self):
        """Get source files from data directory."""
        self._src_data_dir.mkdir(exist_ok=True, parents=True)
        self._src_files = []
        for item_type in ('concepts', 'rels', 'synonyms'):
            src_file_prefix = f'hemonc_{item_type}_'
            dir_files = [f for f in self._src_data_dir.iterdir()
                         if f.name.startswith(src_file_prefix)]
            if len(dir_files) == 0:
                self._download_data()
                dir_files = [f for f in self._src_data_dir.iterdir()
                             if f.name.startswith(src_file_prefix)]
            self._src_files.append(sorted(dir_files, reverse=True)[0])
        self._version = self._src_files[0].stem.split('_', 2)[-1]

    def _load_meta(self):
        """Add HemOnc metadata."""
        meta = {
            'data_license': 'CC BY 4.0',
            'data_license_url': 'https://creativecommons.org/licenses/by/4.0/legalcode',  # noqa: E501
            'version': self._version,
            'data_url': 'https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/DVN/9CY9C6',  # noqa: E501
            'rdp_url': None,
            'data_license_attributes': {
                'non_commercial': False,
                'share_alike': False,
                'attribution': True,
            },
        }
        assert SourceMeta(**meta)
        meta['src_name'] = SourceName.HEMONC.value
        self.database.metadata.put_item(Item=meta)

    def _get_concepts(self) -> Tuple[Dict, Dict, Dict]:
        """Get therapy, brand name, and disease concepts from concepts file.
        :return: Tuple of dicts mapping ID to object for each type of concept
        """
        therapies = {}  # hemonc id -> record
        brand_names = {}  # hemonc id -> brand name
        conditions = {}  # hemonc id -> condition name

        concepts_file = open(self._src_files[0], 'r')
        concepts_reader = csv.reader(concepts_file)
        next(concepts_reader)  # skip header
        for row in concepts_reader:
            if row[6]:
                continue  # skip if deprecated/invalid

            row_type = row[2]
            if row_type == 'Component':
                concept_id = f'{NamespacePrefix.HEMONC.value}:{row[3]}'
                therapies[row[3]] = {
                    'concept_id': concept_id,
                    'label': row[0],
                    'trade_names': [],
                    'aliases': [],
                    'xrefs': [],
                }
            elif row_type == 'Brand Name':
                brand_names[row[3]] = row[0]
            elif row_type == 'Condition':
                conditions[row[3]] = row[0]
        concepts_file.close()

        return therapies, brand_names, conditions

    @staticmethod
    def _id_to_yr(hemonc_id: str) -> int:
        """Get year from HemOnc ID corresponding to year concept.
        :param str hemonc_id: HemOnc ID to get year for
        :return: int representing year. Raises TypeError if HemOnc ID not
            valid.
        """
        id_int = int(hemonc_id)
        if id_int == 780:
            return 9999
        elif id_int == 48349:
            return 2020
        elif id_int == 5963:
            return 2021
        elif id_int < 699 or id_int > 780:
            raise TypeError("ID not a valid HemOnc year concept")
        else:
            return id_int + 1240

    def _get_rels(self, therapies: Dict, brand_names: Dict,
                  conditions: Dict) -> Dict:
        """Gather relations to provide associations between therapies, brand
        names, and conditions.

        :param dict therapies: mapping from IDs to therapy concepts
        :param dict brand_names: mapping from IDs to therapy brand names
        :param dict conditions: mapping from IDs to disease conditions
        :return: therapies dict updated with brand names and conditions
        """
        rels_file = open(self._src_files[1], 'r')
        rels_reader = csv.reader(rels_file)
        next(rels_reader)  # skip header

        for row in rels_reader:
            rel_type = row[4]
            hemonc_id = row[0]
            record = therapies.get(hemonc_id)

            if record is None:
                continue  # skip non-drug items

            if rel_type == "Maps to":
                src_raw = row[3]
                if src_raw == "RxNorm":
                    xref = f'{NamespacePrefix.RXNORM.value}:{row[1]}'
                    record['xrefs'].append(xref)
                elif src_raw == "RxNorm Extension":
                    continue  # skip
                else:
                    logger.warning(f"Unrecognized `Maps To` source: {src_raw}")

            elif rel_type == "Has brand name":
                record['trade_names'].append(brand_names[row[1]])

            elif rel_type == "Was FDA approved yr":
                try:
                    year = self._id_to_yr(row[1])
                except TypeError:
                    logger.error(f"Failed parse of FDA approval year ID "
                                 f"{row[1]} for HemOnc ID {row[0]}")
                    continue
                if year == 9999:
                    logger.warning(f"HemOnc ID {row[0]} has FDA approval year"
                                   f" 9999")
                record['approval_status'] = ApprovalStatus.APPROVED
                if 'approval_year' in record:
                    record['approval_year'].append(year)
                else:
                    record['approval_year'] = [year]

            elif rel_type == "Has FDA indication":
                label = conditions[row[1]]
                norm_response = self.disease_normalizer.search_groups(label)
                if norm_response['match_type'] > 0:
                    ncit_id = norm_response['value_object_descriptor']['value']['id']  # noqa: E501
                else:
                    ncit_id = ""
                    logger.warning(f'Normalization of condition id: {row[1]}'
                                   f', {label}, failed.')
                hemonc_concept_id = f"{NamespacePrefix.HEMONC.value}:{row[1]}"  # noqa: E501
                indication = [hemonc_concept_id, label, ncit_id]
                if 'fda_indication' in record:
                    record['fda_indication'].append(indication)
                else:
                    record['fda_indication'] = [indication]

        rels_file.close()
        return therapies

    def _get_synonyms(self, therapies: Dict) -> Dict:
        """Gather synonym entries and associate with therapy concepts.

        :param dict therapies: mapping of IDs to therapy objects
        :return: therapies dict with synonyms added as aliases
        """
        synonyms_file = open(self._src_files[2], 'r')
        synonyms_reader = csv.reader(synonyms_file)
        next(synonyms_reader)
        for row in synonyms_reader:
            therapy_code = row[1]
            if therapy_code in therapies:
                therapy = therapies[therapy_code]
                alias = row[0]
                if alias != therapy.get('label'):
                    therapies[therapy_code]['aliases'].append(row[0])
        synonyms_file.close()
        return therapies

    def _transform_data(self):
        """Prepare dataset for loading into normalizer database."""
        therapies, brand_names, conditions = self._get_concepts()
        therapies = self._get_rels(therapies, brand_names, conditions)
        therapies = self._get_synonyms(therapies)

        for therapy in therapies.values():
            self._load_therapy(therapy)



"""Module to load and init namespace at package level"""
from .wikidata import Wikidata  # noqa: F401
from .chembl import ChEMBL  # noqa: F401
from .ncit import NCIt  # noqa: F401
from .drugbank import DrugBank  # noqa: F401
from .chemidplus import ChemIDplus  # noqa: F401
from .rxnorm import RxNorm  # noqa: F401
from .hemonc import HemOnc  # noqa: F401
from .merge import Merge  # noqa: F401



"""This module contains data models for representing VICC normalized
therapy records.
"""
from typing import List, Optional, Dict, Union, Any, Type
from pydantic import BaseModel, StrictBool
from enum import Enum, IntEnum
from datetime import datetime


class Therapy(BaseModel):
    """A procedure or substance used in the treatment of a disease."""

    label: str
    concept_id: str
    aliases: Optional[List[str]]
    xrefs: Optional[List[str]]
    associated_with: Optional[List[str]]

    class Config:
        """Configure model"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['Therapy']) -> None:
            """Configure OpenAPI schema"""
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)


class ApprovalStatus(str, Enum):
    """Define string constraints for approval status attribute."""

    WITHDRAWN = "withdrawn"
    APPROVED = "approved"
    INVESTIGATIONAL = "investigational"


class PhaseEnum(IntEnum):
    """An enumerated drug development phase type."""

    preclinical = 0
    phase_i_trials = 1
    phase_ii_trials = 2
    phase_iii_trials = 3
    approved = 4


class HasIndication(BaseModel):
    """Data regarding FDA indication. Currently specific to HemOnc.org data."""

    disease_id: str
    disease_label: str
    normalized_disease_id: Optional[str]

    class Config:
        """Configure Drug class"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['Drug']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = [
                {
                    "disease_id": "hemonc:671",
                    "disease_label": "Testicular cancer",
                    "normalized_disease_id": "ncit:C7251"
                },
                {
                    "disease_id": "hemonc:645",
                    "disease_label": "Ovarian cancer",
                    "normalized_disease_id": "ncit:C7431"
                },
                {
                    "disease_id": "hemonc:569",
                    "disease_label": "Bladder cancer",
                    "normalized_disease_id": "ncit:C9334"
                }
            ]


class Drug(Therapy):
    """A pharmacologic substance used to treat a medical condition."""

    approval_status: Optional[ApprovalStatus]
    approval_year: Optional[List[int]]
    has_indication: Optional[List[HasIndication]]
    trade_names: Optional[List[str]]
    label: Optional[str]

    class Config:
        """Configure Drug class"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['Drug']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                "label": "CISPLATIN",
                "concept_id": "chembl:CHEMBL11359",
                "aliases": [
                    "Cisplatin",
                    "Cis-Platinum II",
                    "Cisplatinum",
                    "cis-diamminedichloroplatinum(II)",
                    "CIS-DDP",
                    "INT-230-6 COMPONENT CISPLATIN",
                    "INT230-6 COMPONENT CISPLATIN",
                    "NSC-119875",
                    "Platinol",
                    "Platinol-Aq"
                ],
                "xrefs": [],
                "associated_with": None,
                "approval_status": "approved",
                "approval_year": [],
                "has_indication": [],
                "trade_names": ["PLATINOL", "PLATINOL-AQ", "CISPLATIN"]
            }


class DrugGroup(Therapy):
    """A grouping of drugs based on common pharmacological attributes."""

    description: str
    type_identifier: str
    drugs: List[Drug]


class MatchType(IntEnum):
    """Define string constraints for use in Match Type attributes.

    Concept_ID=100; Label=80; Trade Name=80; Alias=60; Fuzzy=20; No Match=0
    """

    CONCEPT_ID = 100
    LABEL = 80
    TRADE_NAME = 80
    ALIAS = 60
    XREF = 60
    ASSOCIATED_WITH = 60
    FUZZY_MATCH = 20
    NO_MATCH = 0


class SourcePriority(IntEnum):
    """Define constraints for Source Priority Rankings."""

    RXNORM = 1
    NCIT = 2
    HEMONC = 3
    DRUGBANK = 4
    CHEMIDPLUS = 5
    WIKIDATA = 10


class SourceName(str, Enum):
    """Define string constraints to ensure consistent capitalization."""

    WIKIDATA = "Wikidata"
    CHEMBL = "ChEMBL"
    NCIT = "NCIt"
    DRUGBANK = "DrugBank"
    CHEMIDPLUS = "ChemIDplus"
    RXNORM = "RxNorm"
    HEMONC = "HemOnc"


class SourceIDAfterNamespace(Enum):
    """Define string constraints after namespace."""

    WIKIDATA = "Q"
    CHEMBL = "CHEMBL"
    DRUGBANK = "DB"
    NCIT = "C"
    CHEMIDPLUS = ""
    RXNORM = ""
    HEMONC = ""


class ProhibitedSources(Enum):
    """Define constraints for sources that are prohibited in normalize
    endpoint.
    """

    CHEMBL = SourceName.CHEMBL.value


class NamespacePrefix(Enum):
    """Define string constraints for namespace prefixes on concept IDs."""

    CHEMIDPLUS = "chemidplus"
    CASREGISTRY = CHEMIDPLUS
    PUBCHEMCOMPOUND = "pubchem.compound"
    PUBCHEMSUBSTANCE = "pubchem.substance"
    CHEMBL = "chembl"
    RXNORM = "rxcui"
    DRUGBANK = "drugbank"
    WIKIDATA = "wikidata"
    HEMONC = "hemonc"
    NCIT = "ncit"
    FDA = "fda"
    ISO = "iso"
    UMLS = "umls"
    CHEBI = "chebi"
    KEGGCOMPOUND = "kegg.compound"
    KEGGDRUG = "kegg.drug"
    BINDINGDB = "bindingdb"
    PHARMGKB = "pharmgkb.drug"
    CHEMSPIDER = "chemspider"
    ZINC = "zinc"
    PDB = "pdb"
    THERAPEUTICTARGETSDB = "ttd"
    IUPHAR = "iuphar"
    GUIDETOPHARMACOLOGY = "gtopdb"
    INCHIKEY = "inchikey"
    UNII = "unii"
    ATC = "atc"  # Anatomical Therapeutic Chemical Classification System
    CVX = "cvx"  # Vaccines Administered
    MMSL = "mmsl"  # Multum MediSource Lexicon
    MSH = "mesh"  # Medical Subject Headings
    MTHCMSFRF = "mthcmsfrf"  # CMS Formulary Reference File
    MTHSPL = "mthspl"  # FDA Structured Product Labels
    USP = "usp"  # USP Compendial Nomenclature
    VANDF = "vandf"  # Veterans Health Administration National Drug File


class DataLicenseAttributes(BaseModel):
    """Define constraints for data license attributes."""

    non_commercial: StrictBool
    share_alike: StrictBool
    attribution: StrictBool


class SourceMeta(BaseModel):
    """Metadata for a given source to return in response object."""

    data_license: str
    data_license_url: str
    version: str
    data_url: Optional[str]
    rdp_url: Optional[str]
    data_license_attributes: Dict[str, StrictBool]

    class Config:
        """Configure OpenAPI schema"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['SourceMeta']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                'data_license': 'CC BY-SA 3.0',
                'data_license_url':
                    'https://creativecommons.org/licenses/by-sa/3.0/',
                'version': '27',
                'data_url':
                    'http://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_27/',  # noqa: E501
                'rdp_url': 'http://reusabledata.org/chembl.html',
                'data_license_attributes': {
                    'non_commercial': False,
                    'share_alike': True,
                    'attribution': True
                }
            }


class MatchesKeyed(BaseModel):
    """Container for matching information from an individual source.
    Used when matches are requested as an object, not an array.
    """

    match_type: MatchType
    records: List[Drug]
    source_meta_: SourceMeta

    class Config:
        """Configure OpenAPI schema"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['MatchesKeyed']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                'match_type': 0,
                'records': [],
                'source_meta_': {
                    'data_license': 'CC BY-SA 3.0',
                    'data_license_url':
                        'https://creativecommons.org/licenses/by-sa/3.0/',
                    'version': '27',
                    'data_url':
                        'http://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_27/',  # noqa: E501
                    'rdp_url': 'http://reusabledata.org/chembl.html',
                    'data_license_attributes': {
                        'non_commercial': False,
                        'share_alike': True,
                        'attribution': True
                    }
                },
            }


class MatchesListed(BaseModel):
    """Container for matching information from an individual source.
    Used when matches are requested as an array, not an object.
    """

    source: SourceName
    match_type: MatchType
    records: List[Drug]
    source_meta_: SourceMeta

    class Config:
        """Configure openAPI schema"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['MatchesListed']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                'normalizer': 'ChEMBL',
                'match_type': 0,
                'records': [],
                'source_meta_': {
                    'data_license': 'CC BY-SA 3.0',
                    'data_license_url':
                        'https://creativecommons.org/licenses/by-sa/3.0/',
                    'version': '27',
                    'data_url':
                        'http://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_27/',  # noqa: E501
                    'rdp_url': 'http://reusabledata.org/chembl.html',
                    'data_license_attributes': {
                        'non_commercial': False,
                        'share_alike': True,
                        'attribution': True
                    }
                },
            }


class DiseaseDescriptor(BaseModel):
    """VOD class for diseases. Used within the FDAStatus extension in the
    main therapy VOD.
    """

    id: str
    type = "DiseaseDescriptor"
    value: Optional[Dict]
    label: str


class FDAStatus(BaseModel):
    """VOD Extension class for FDA status/indication attributes."""

    approval_status: Optional[ApprovalStatus]
    approval_year: Optional[List[int]]
    has_indication: Optional[List[Union[Dict, DiseaseDescriptor]]]


class Extension(BaseModel):
    """Value Object Descriptor Extension class."""

    type = "Extension"
    name: str
    value: Union[bool, List[str], FDAStatus]


class TherapyDescriptor(BaseModel):
    """Therapy VOD class. Provide additional Extension classes for trade_names
    and references to non-normalized sources as well as FDA approval data.
    """

    id: str
    type = "TherapyDescriptor"
    value: Any
    label: str
    xrefs: Optional[List[str]]
    alternate_labels: Optional[List[str]]
    extensions: Optional[List[Extension]]


class ServiceMeta(BaseModel):
    """Metadata regarding the therapy-normalization service."""

    name = "thera-py"
    version: str
    response_datetime: datetime
    url: str

    class Config:
        """Configure OpenAPI schema"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['SourceMeta']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                'name': 'thera-py',
                'version': '0.1.0',
                'response_datetime': '2021-04-05T16:44:15.367831',
                'url': 'https://github.com/cancervariants/therapy-normalization'  # noqa: E501
            }


class NormalizationService(BaseModel):
    """Response containing one or more merged records and source data."""

    query: str
    warnings: Optional[Dict]
    match_type: MatchType
    value_object_descriptor: Optional[TherapyDescriptor]
    source_meta_: Optional[Dict[SourceName, SourceMeta]]
    service_meta_: ServiceMeta

    class Config:
        """Configure OpenAPI schema"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['NormalizationService']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                "query": "cisplatin",
                "warnings": None,
                "match_type": 80,
                "value_object_descriptor": {
                    "id": "normalize.therapy:cisplatin",
                    "type": "TherapyDescriptor",
                    "value": {
                        "type": "Therapy",
                        "id": "rxcui:2555"
                    },
                    "label": "cisplatin",
                    "xrefs": [
                        "ncit:C376", "chemidplus:15663-27-1",
                        "wikidata:Q412415"
                    ],
                    "alternate_labels": [
                        "CIS-DDP", "cis Platinum", "DDP",
                        "Dichlorodiammineplatinum",
                        "1,2-Diaminocyclohexaneplatinum II citrate",
                        "CISplatin",
                        "cis Diamminedichloroplatinum",
                        "CDDP",
                        "Diamminodichloride, Platinum",
                        "cis-Dichlorodiammineplatinum(II)",
                        "cis-Platinum",
                        "cis-diamminedichloroplatinum(II)",
                        "cis-Diamminedichloroplatinum(II)",
                        "Cis-DDP",
                        "cis-Diamminedichloroplatinum",
                        "cis-Diaminedichloroplatinum",
                        "Platinol-AQ", "Platinol",
                        "Platinum Diamminodichloride"
                    ],
                    "extensions": [
                        {
                            "type": "Extension",
                            "name": "trade_names",
                            "value": [
                                "Platinol", "Cisplatin"
                            ]
                        },
                        {
                            "type": "Extension",
                            "name": "associated_with",
                            "value": [
                                "atc:L01XA01",
                                "mmsl:4456",
                                "chebi:CHEBI:27899",
                                "pubchem.compound:5702198",
                                "umls:C0008838",
                                "usp:m17910",
                                "fda:Q20Q21Q62J",
                                "mmsl:d00195",
                                "mthspl:Q20Q21Q62J",
                                "mmsl:31747",
                                "mesh:D002945",
                                "vandf:4018139"
                            ]
                        }
                    ]
                },
                "source_meta_": {
                    "RxNorm": {
                        "data_license": "UMLS Metathesaurus",
                        "data_license_url": "https://www.nlm.nih.gov/research/umls/rxnorm/docs/termsofservice.html",  # noqa: E501
                        "version": "20210104",
                        "data_url": "https://www.nlm.nih.gov/research/umls/rxnorm/docs/rxnormfiles.html",  # noqa: E501
                        "rdp_url": None,
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": True,
                            "share_alike": False
                        }
                    },
                    "NCIt": {
                        "data_license": "CC BY 4.0",
                        "data_license_url": "https://creativecommons.org/licenses/by/4.0/legalcode",  # noqa: E501
                        "version": "20.09d",
                        "data_url": "https://evs.nci.nih.gov/ftp1/NCI_Thesaurus/archive/20.09d_Release/",  # noqa: E501
                        "rdp_url": "http://reusabledata.org/ncit.html",
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": True,
                            "share_alike": False
                        }
                    },
                    "ChemIDplus": {
                        "data_license": "custom",
                        "data_license_url": "https://www.nlm.nih.gov/databases/download/terms_and_conditions.html",  # noqa: E501
                        "version": "20200327", "data_url": "ftp://ftp.nlm.nih.gov/nlmdata/.chemidlease/",  # noqa: E501
                        "rdp_url": None,
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": True,
                            "share_alike": False
                        }
                    },
                    "Wikidata": {
                        "data_license": "CC0 1.0",
                        "data_license_url": "https://creativecommons.org/publicdomain/zero/1.0/",  # noqa: E501
                        "version": "20200812",
                        "data_url": None,
                        "rdp_url": None,
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": False,
                            "share_alike": False
                        }
                    }
                },
                "service_meta_": {
                    'name': 'thera-py',
                    'version': '0.1.0',
                    'response_datetime': '2021-04-05T16:44:15.367831',
                    'url': 'https://github.com/cancervariants/therapy-normalization'  # noqa: E501
                }
            }


class SearchService(BaseModel):
    """Core response schema containing matches for each source"""

    query: str
    warnings: Optional[Dict]
    source_matches: Union[Dict[SourceName, MatchesKeyed], List[MatchesListed]]
    service_meta_: ServiceMeta

    class Config:
        """Enables orm_mode"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['SearchService']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                "query": "cisplatin",
                "warnings": None,
                "source_matches": [
                    {
                        "source": "ChemIDplus",
                        "match_type": 80,
                        "records": [
                            {
                                "label": "Cisplatin",
                                "concept_id": "chemidplus:15663-27-1",
                                "aliases": [
                                    "cis-Diaminedichloroplatinum",
                                    "1,2-Diaminocyclohexaneplatinum II citrate"
                                ],
                                "xrefs": ["drugbank:DB00515"],
                                "associated_with": ["fda:Q20Q21Q62J"],
                                "approval_status": None,
                                "trade_names": []
                            }
                        ],
                        "source_meta_": {
                            "data_license": "custom",
                            "data_license_url": "https://www.nlm.nih.gov/databases/download/terms_and_conditions.html",  # noqa: E501
                            "version": "20210204",
                            "data_url": "ftp://ftp.nlm.nih.gov/nlmdata/.chemidlease/",  # noqa: E501
                            "rdp_url": None,
                            "data_license_attributes": {
                                "non_commercial": False,
                                "attribution": True,
                                "share_alike": False
                            }
                        }
                    },
                    {
                        "source": "RxNorm",
                        "match_type": 80,
                        "records": [
                            {
                                "label": "cisplatin",
                                "concept_id": "rxcui:2555",
                                "aliases": [
                                    "cis-Dichlorodiammineplatinum(II)",
                                    "Platinum Diamminodichloride",
                                    "cis Diamminedichloroplatinum",
                                    "cis-diamminedichloroplatinum(II)",
                                    "cis-Diamminedichloroplatinum",
                                    "cis Platinum",
                                    "CDDP",
                                    "Dichlorodiammineplatinum",
                                    "cis-Platinum",
                                    "CISplatin",
                                    "cis-Diamminedichloroplatinum(II)",
                                    "Cis-DDP",
                                    "DDP",
                                    "Diamminodichloride, Platinum"
                                ],
                                "xrefs": [
                                    "drugbank:DB00515",
                                    "drugbank:DB12117"
                                ],
                                "associated_with": [
                                    "usp:m17910",
                                    "vandf:4018139",
                                    "mesh:D002945",
                                    "mthspl:Q20Q21Q62J",
                                    "mmsl:d00195",
                                    "atc:L01XA01",
                                    "mmsl:31747",
                                    "mmsl:4456"
                                ],
                                "approval_status": "approved",
                                "trade_names": [
                                    "Cisplatin",
                                    "Platinol"
                                ]
                            }
                        ],
                        "source_meta_": {
                            "data_license": "UMLS Metathesaurus",
                            "data_license_url": "https://www.nlm.nih.gov/research/umls/rxnorm/docs/termsofservice.html",  # noqa: E501
                            "version": "20210104",
                            "data_url": "https://www.nlm.nih.gov/research/umls/rxnorm/docs/rxnormfiles.html",  # noqa: E501
                            "rdp_url": None,
                            "data_license_attributes": {
                                "non_commercial": False,
                                "attribution": True,
                                "share_alike": False
                            }
                        }
                    },
                    {
                        "source": "NCIt",
                        "match_type": 80,
                        "records": [
                            {
                                "label": "Cisplatin",
                                "concept_id": "ncit:C376",
                                "aliases": [],
                                "xrefs": ["chemidplus:15663-27-1"],
                                "associated_with": [
                                    "umls:C0008838",
                                    "fda:Q20Q21Q62J",
                                    "chebi:CHEBI:27899"
                                ],
                                "approval_status": None,
                                "trade_names": []
                            }
                        ],
                        "source_meta_": {
                            "data_license": "CC BY 4.0",
                            "data_license_url": "https://creativecommons.org/licenses/by/4.0/legalcode",  # noqa: E501
                            "version": "20.09d",
                            "data_url": "https://evs.nci.nih.gov/ftp1/NCI_Thesaurus/archive/2020/20.09d_Release/",  # noqa: E501
                            "rdp_url": "http://reusabledata.org/ncit.html",
                            "data_license_attributes": {
                                "non_commercial": False,
                                "attribution": True,
                                "share_alike": False
                            }
                        }
                    },
                    {
                        "source": "Wikidata",
                        "match_type": 80,
                        "records": [
                            {
                                "label": "cisplatin",
                                "concept_id": "wikidata:Q412415",
                                "aliases": [
                                    "Platinol",
                                    "cis-diamminedichloroplatinum(II)",
                                    "CDDP",
                                    "Cis-DDP",
                                    "CIS-DDP",
                                    "Platinol-AQ"
                                ],
                                "xrefs": [
                                    "chemidplus:15663-27-1",
                                    "chembl:CHEMBL11359",
                                    "rxcui:2555",
                                    "drugbank:DB00515"
                                ],
                                "associated_with": [
                                    "pubchem.compound:5702198"
                                ],
                                "approval_status": None,
                                "trade_names": []
                            }
                        ],
                        "source_meta_": {
                            "data_license": "CC0 1.0",
                            "data_license_url": "https://creativecommons.org/publicdomain/zero/1.0/",  # noqa: E501
                            "version": "20210331",
                            "data_url": None,
                            "rdp_url": None,
                            "data_license_attributes": {
                                "non_commercial": False,
                                "attribution": False,
                                "share_alike": False
                            }
                        }
                    }
                ],
                "service_meta_": {
                    'name': 'thera-py',
                    'version': '0.1.0',
                    'response_datetime': '2021-04-05T16:44:15.367831',
                    'url': 'https://github.com/cancervariants/therapy-normalization'  # noqa: E501
                }
            }



"""This module defines the RxNorm ETL methods.

"This product uses publicly available data courtesy of the U.S. National
Library of Medicine (NLM), National Institutes of Health, Department of Health
 and Human Services; NLM is not responsible for the product and does not
 endorse or recommend this or any other product."
"""
from .base import Base
from therapy import PROJECT_ROOT, DownloadException, XREF_SOURCES, \
    ASSOC_WITH_SOURCES
import therapy
from therapy.database import Database
from therapy.schemas import SourceName, NamespacePrefix, SourceMeta, Drug, \
    ApprovalStatus
import csv
import datetime
import logging
import botocore
from typing import List
from os import environ, remove
import subprocess
import shutil
import zipfile
import re
import yaml
from pathlib import Path

logger = logging.getLogger('therapy')
logger.setLevel(logging.DEBUG)

# Designated Alias, Designated Syn, Tall Man Syn, Machine permutation
# Generic Drug Name, Designated Preferred Name, Preferred Entry Term,
# Clinical Drug, Entry Term, Rxnorm Preferred
ALIASES = ['SYN', 'SY', 'TMSY', 'PM',
           'GN', 'PT', 'PEP', 'CD', 'ET', 'RXN_PT']

# Fully-specified drug brand name that can be prescribed
# Fully-specified drug brand name that can not be prescribed,
# Semantic branded drug
TRADE_NAMES = ['BD', 'BN', 'SBD']

# Allowed rxnorm xrefs that have Source Level Restriction 0 or 1
RXNORM_XREFS = ['ATC', 'CVX', 'DRUGBANK', 'MMSL', 'MSH', 'MTHCMSFRF', 'MTHSPL',
                'RXNORM', 'USP', 'VANDF']


class RxNorm(Base):
    """Extract, transform, and load the RxNorm source."""

    def __init__(self,
                 database: Database,
                 data_url='https://www.nlm.nih.gov/research/umls/'
                          'rxnorm/docs/rxnormfiles.html'):
        """Initialize the RxNorm ETL class.

        :param Database database: Access to DynamoDB.
        :param str data_url: URL to RxNorm data files.
        """
        self._data_url = data_url
        self.database = database
        self._added_ids = []

    def perform_etl(self) -> List[str]:
        """Load the RxNorm source into database.

        :return: List of concept IDs which were successfully processed and
            uploaded.
        """
        self._extract_data()
        self._load_meta()
        self._transform_data()
        return self._added_ids

    def _extract_data(self, *args, **kwargs):
        """Extract data from the RxNorm source."""
        logger.info('Extracting RxNorm...')
        if 'data_path' in kwargs:
            self._data_src = kwargs['data_path']
        else:
            rxn_dir = PROJECT_ROOT / 'data' / 'rxnorm'
            rxn_dir.mkdir(exist_ok=True, parents=True)
            rxn_files = list(rxn_dir.iterdir())
            if len(rxn_files) == 0:
                self._download_data(rxn_dir)

            # file might already exist
            files = sorted([fn for fn in rxn_dir.iterdir() if fn.name.
                           startswith('rxnorm_') and fn.name.endswith('.RRF')],
                           reverse=True)
            file_found = False
            for file in files:
                version = str(file).split('_')[-1].split('.')[0]
                try:
                    datetime.datetime.strptime(version, '%Y%m%d')
                    self._data_src = file
                    self._version = version
                    file_found = True
                    break
                except ValueError:
                    pass
            if not file_found:
                self._download_data(rxn_dir)
            self._create_drug_form_yaml(rxn_dir)
        logger.info('Successfully extracted RxNorm.')

    def _download_data(self, rxn_dir):
        """Download RxNorm data file.

        :param Path rxn_dir: Path to RxNorm data directory.
        """
        logger.info('Downloading RxNorm...')
        if 'RXNORM_API_KEY' in environ.keys():
            uri = 'https://download.nlm.nih.gov/umls/kss/' \
                  'rxnorm/RxNorm_full_current.zip'

            rxnorm_path = str(Path(therapy.__file__).resolve().parents[0] / 'data' / 'rxnorm' / 'RxNorm_full_current.zip')  # noqa: E501
            environ['RXNORM_PATH'] = rxnorm_path

            # Source:
            # https://documentation.uts.nlm.nih.gov/automating-downloads.html
            subprocess.call(['bash', f'{PROJECT_ROOT}/etl/'
                                     f'rxnorm_download.sh', uri])

            with zipfile.ZipFile(rxnorm_path, 'r') as zf:
                zf.extractall(rxn_dir)

            remove(rxnorm_path)
            shutil.rmtree(rxn_dir / 'prescribe')
            shutil.rmtree(rxn_dir / 'scripts')

            readme = sorted([fn for fn in rxn_dir.iterdir() if fn.name.
                            startswith('Readme')])[0]

            # get version
            version = str(readme).split('/')[-1].split('.')[0].split('_')[-1]

            self._version = datetime.datetime.strptime(
                version, '%m%d%Y').strftime('%Y%m%d')
            remove(readme)

            temp_file = rxn_dir / 'rrf' / 'RXNCONSO.RRF'
            self._data_src = rxn_dir / f"rxnorm_{self._version}.RRF"
            shutil.move(temp_file, self._data_src)
            shutil.rmtree(rxn_dir / 'rrf')
        else:
            logger.error('Could not find RXNORM_API_KEY in environment '
                         'variables.')
            raise DownloadException("RXNORM_API_KEY not found.")

    def _create_drug_form_yaml(self, rxn_dir):
        """Create a YAML file containing RxNorm drug form values.

        :param Path rxn_dir: Path to RxNorm data directory.
        """
        self._drug_forms_file = rxn_dir / 'drug_forms.yaml'
        if not self._drug_forms_file.exists():
            dfs = list()
            with open(self._data_src) as f:
                data = csv.reader(f, delimiter='|')
                for row in data:
                    if row[12] == 'DF' and row[11] == 'RXNORM':
                        if row[14] not in dfs:
                            dfs.append(row[14])

            with open(self._drug_forms_file, 'w') as file:
                yaml.dump(dfs, file)

    def _transform_data(self):
        """Transform the RxNorm source."""
        with open(self._drug_forms_file, 'r') as file:
            drug_forms = yaml.safe_load(file)

        with open(self._data_src) as f:
            rff_data = csv.reader(f, delimiter='|')
            ingredient_brands = dict()  # Link ingredient to brand
            precise_ingredient = dict()  # Link precise ingredient to get brand
            data = dict()  # Transformed therapy records
            sbdfs = dict()  # Link ingredient to brand
            brands = dict()  # Get RXNORM|BN to concept_id
            for row in rff_data:
                if row[11] in RXNORM_XREFS:
                    concept_id = f"{NamespacePrefix.RXNORM.value}:{row[0]}"
                    if row[12] == 'BN' and row[11] == 'RXNORM':
                        brands[row[14]] = concept_id
                    if row[12] == 'SBDC' and row[11] == 'RXNORM':
                        # Semantic Branded Drug Component
                        self._get_brands(row, ingredient_brands)
                    else:
                        if concept_id not in data.keys():
                            params = dict()
                            params['concept_id'] = concept_id
                            self._add_str_field(params, row,
                                                precise_ingredient,
                                                drug_forms, sbdfs)
                            self._add_xref_assoc(params, row)
                            data[concept_id] = params
                        else:
                            # Concept already created
                            params = data[concept_id]
                            self._add_str_field(params, row,
                                                precise_ingredient,
                                                drug_forms, sbdfs)
                            self._add_xref_assoc(params, row)

            with self.database.therapies.batch_writer() as batch:
                for key, value in data.items():
                    if 'label' in value:
                        self._get_trade_names(value, precise_ingredient,
                                              ingredient_brands, sbdfs)
                        self._load_brand_concepts(value, brands, batch)

                        params = Drug(
                            concept_id=value['concept_id'],
                            label=value['label'] if 'label' in value else None,
                            approval_status=value[
                                'approval_status'] if 'approval_status'
                                                      in value else None,
                            aliases=value['aliases'] if 'aliases' in
                                                        value else [],
                            xrefs=value['xrefs'] if 'xrefs' in value else [],
                            associated_with=value[
                                'associated_with'] if 'associated_with' in
                                                      value else [],
                            trade_names=value['trade_names'] if 'trade_names'
                                                                in value
                                                                else []
                        )
                        self._load_therapy(params, batch)

    def _load_therapy(self, params, batch):
        """Load therapy record into the database.

        :param Drug params: A Therapy object.
        :param BatchWriter batch: Object to write data to DynamoDB.
        """
        params = dict(params)
        for label_type in ['label', 'aliases', 'xrefs', 'associated_with',
                           'trade_names']:
            if not params[label_type]:
                del params[label_type]

        for attr in ['trade_names', 'aliases']:
            # Remove duplicates
            if attr in params:
                if 'label' in params:
                    params[attr] = list(set(params[attr]) - {params['label']})

                if attr == 'aliases' and 'trade_names' in params:
                    params[attr] = \
                        list(set(params[attr]) - set(params['trade_names']))

                # 20 cutoff for aliases / trade names
                if len(params[attr]) > 20:
                    logger.debug(f"{params['concept_id']} has more than 20"
                                 f" {attr}.")
                    del params[attr]

        self._load_label_types(params, batch)
        params['src_name'] = SourceName.RXNORM.value
        params['label_and_type'] = f"{params['concept_id'].lower()}##identity"
        params['item_type'] = 'identity'
        try:
            batch.put_item(Item=params)
            self._added_ids.append(params['concept_id'])
        except botocore.exceptions.ClientError:
            if (len((params['label_and_type']).encode('utf-8')) >= 2048) or \
                    (len((params['concept_id']).encode('utf-8')) >= 1024):
                logger.info(f"{params['concept_id']}: An error occurred "
                            "(ValidationException) when calling the "
                            "BatchWriteItem operation: Hash primary key "
                            "values must be under 2048 bytes, and range"
                            " primary key values must be under 1024"
                            " bytes.")

    def _load_label_types(self, params, batch):
        """Insert aliases, trade_names, and label data into the database.

        :param dict params: A transformed therapy record.
        :param BatchWriter batch: Object to write data to DynamoDB
        """
        if 'label' in params:
            self._load_label_type(params, batch, 'label', 'label')
        if 'trade_names' in params:
            self._load_label_type(params, batch, 'trade_name', 'trade_names')
        if 'aliases' in params:
            self._load_label_type(params, batch, 'alias', 'aliases')
        if 'xrefs' in params:
            self._load_label_type(params, batch, 'xref', 'xrefs')
        if 'associated_with' in params:
            self._load_label_type(params, batch, 'associated_with',
                                  'associated_with')

    def _load_label_type(self, params, batch, label_type_sing, label_type_pl):
        """Insert alias, trade_name, or label data into the database.

        :param dict params: A transformed therapy record.
        :param BatchWriter batch: Object to write data to DynamoDB
        :param str label_type_sing: The singular label type
        :param str label_type_pl: The plural label type
        """
        if isinstance(params[label_type_pl], list):
            terms = {t.casefold(): t for t in params[label_type_pl]}.values()
        else:
            terms = [params[label_type_pl]]
        for t in terms:
            t = {
                'label_and_type': f"{t.lower()}##{label_type_sing}",
                'concept_id': f"{params['concept_id'].lower()}",
                'src_name': SourceName.RXNORM.value,
                'item_type': label_type_sing
            }
            try:
                batch.put_item(Item=t)
            except botocore.exceptions.ClientError:
                if (len((t['label_and_type']).encode('utf-8')) >= 2048) or \
                        (len((t['concept_id']).encode('utf-8')) >= 1024):
                    logger.info(f"{params['concept_id']}: An error occurred "
                                "(ValidationException) when calling the "
                                "BatchWriteItem operation: Hash primary key "
                                "values must be under 2048 bytes, and range"
                                " primary key values must be under 1024"
                                " bytes.")
        return terms

    def _get_brands(self, row, ingredient_brands):
        """Add ingredient and brand to ingredient_brands.

        :param list row: A row in the RxNorm data file.
        :param dict ingredient_brands: Store brands for each ingredient
        """
        # SBDC: Ingredient(s) + Strength + [Brand Name]
        term = row[14]
        ingredients_brand = \
            re.sub(r"(\d*)(\d*\.)?\d+ (MG|UNT|ML)?(/(ML|HR|MG))?",
                   "", term)
        brand = term.split('[')[-1].split(']')[0]
        ingredients = ingredients_brand.replace(f"[{brand}]", '')
        if '/' in ingredients:
            ingredients = ingredients.split('/')
            for ingredient in ingredients:
                self._add_term(ingredient_brands, brand,
                               ingredient.strip())
        else:
            self._add_term(ingredient_brands, brand,
                           ingredients.strip())

    def _get_trade_names(self, value, precise_ingredient, ingredient_brands,
                         sbdfs):
        """Get trade names for a given ingredient.

        :param dict value: Therapy attributes
        :param dict precise_ingredient: Brand names for precise ingredient
        :param dict ingredient_brands: Brand names for ingredient
        :param dict sbdfs: Brand names for ingredient from SBDF row
        """
        record_label = value['label'].lower()
        labels = [record_label]

        if 'PIN' in value and value['PIN'] \
                in precise_ingredient:
            for pin in precise_ingredient[value['PIN']]:
                labels.append(pin.lower())

        for label in labels:
            trade_names = \
                [val for key, val in ingredient_brands.items()
                 if label == key.lower()]
            trade_names = {val for sublist in trade_names
                           for val in sublist}
            for tn in trade_names:
                self._add_term(value, tn, 'trade_names')

        if record_label in sbdfs:
            for tn in sbdfs[record_label]:
                self._add_term(value, tn, 'trade_names')

    def _load_brand_concepts(self, value, brands, batch):
        """Connect brand names to a concept and load into the database.

        :params dict value: A transformed therapy record
        :params dict brands: Connects brand names to concept records
        :param BatchWriter batch: Object to write data to DynamoDB.
        """
        if 'trade_names' in value:
            for tn in value['trade_names']:
                if brands.get(tn):
                    batch.put_item(Item={
                        'label_and_type':
                            f"{brands.get(tn)}##rx_brand",
                        'concept_id': value['concept_id'],
                        'src_name': SourceName.RXNORM.value,
                        'item_type': 'rx_brand'
                    })

    def _add_str_field(self, params, row, precise_ingredient, drug_forms,
                       sbdfs):
        """Differentiate STR field.

        :param dict params: A transformed therapy record.
        :param list row: A row in the RxNorm data file.
        :param dict precise_ingredient: Precise ingredient information
        :param list drug_forms: RxNorm Drug Form values
        :param dict precise_ingredient: Brand names for precise ingredient
        """
        term = row[14]
        term_type = row[12]
        source = row[11]

        if (term_type == 'IN' or term_type == 'PIN') and source == 'RXNORM':
            params['label'] = term
            if row[17] == '4096':
                params['approval_status'] = ApprovalStatus.APPROVED.value
        elif term_type in ALIASES:
            self._add_term(params, term, 'aliases')
        elif term_type in TRADE_NAMES:
            self._add_term(params, term, 'trade_names')

        if source == 'RXNORM':
            if term_type == 'SBDF':
                brand = term.split('[')[-1].split(']')[0]
                ingredient_strength = term.replace(f"[{brand}]", '')
                for df in drug_forms:
                    if df in ingredient_strength:
                        ingredient = \
                            ingredient_strength.replace(df, '').strip()
                        self._add_term(sbdfs, brand, ingredient.lower())
                        break

        if source == 'MSH':
            if term_type == 'MH':
                # Get ID for accessing precise ingredient
                params['PIN'] = row[13]
            elif term_type == 'PEP':
                self._add_term(precise_ingredient, term, row[13])

    def _add_term(self, params, term, label_type):
        """Add a single term to a therapy record in an associated field.

        :param dict params: A transformed therapy record.
        :param str term: The term to be added
        :param str label_type: The type of term
        """
        if label_type in params and params[label_type]:
            if term not in params[label_type]:
                params[label_type].append(term)
        else:
            params[label_type] = [term]

    def _add_xref_assoc(self, params, row):
        """Add xref or associated_with to therapy.

        :param dict params: A transformed therapy record.
        :param list row: A row in the RxNorm data file.
        """
        if row[11]:
            xref_assoc = row[11].upper()
            if xref_assoc in XREF_SOURCES:
                source_id =\
                    f"{NamespacePrefix[xref_assoc].value}:{row[13]}"
                if source_id != params['concept_id']:
                    # Sometimes concept_id is included in the source field
                    self._add_term(params, source_id, 'xrefs')
            elif xref_assoc in ASSOC_WITH_SOURCES:
                source_id = f"{NamespacePrefix[xref_assoc].value}:{row[13]}"
                self._add_term(params, source_id, 'associated_with')
            else:
                logger.info(f"{xref_assoc} not in NameSpacePrefix.")

    def _load_meta(self):
        """Add RxNorm metadata."""
        meta = SourceMeta(
            data_license='UMLS Metathesaurus',
            data_license_url='https://www.nlm.nih.gov/research/umls/rxnorm/docs/termsofservice.html',  # noqa: E501
            version=self._version,
            data_url=self._data_url,
            rdp_url=None,
            data_license_attributes={
                'non_commercial': False,
                'share_alike': False,
                'attribution': True
            }
        )
        params = dict(meta)
        params['src_name'] = SourceName.RXNORM.value
        self.database.metadata.put_item(Item=params)



"""ETL methods for ChemIDPlus® source.

Courtesy of the U.S. National Library of Medicine.
"""
from .base import Base
from typing import List, Dict
from therapy import PROJECT_ROOT
from therapy.schemas import Drug, NamespacePrefix, SourceMeta, SourceName, \
    DataLicenseAttributes
from pathlib import Path
from ftplib import FTP
import xml.etree.ElementTree as ET
import logging
from boto3.dynamodb.table import BatchWriter
import re


logger = logging.getLogger('therapy')
logger.setLevel(logging.DEBUG)


TAGS_REGEX = r' \[.*\]'


class ChemIDplus(Base):
    """Core ChemIDplus ETL class."""

    def __init__(self,
                 database,
                 data_path: Path = PROJECT_ROOT / 'data' / 'chemidplus',
                 src_server: str = 'ftp.nlm.nih.gov',
                 src_dir_path: str = 'nlmdata/.chemidlease/',
                 src_fname: str = 'CurrentChemID.xml'):
        """Initialize class instance.

        :param therapy.database.Database database: database instance to use
        :param Path data_path: path to chemidplus subdirectory in application
            data folder
        :param str src_server: The NLM domain
        :param str src_dir_path: The directory to the chemidplus release
        :param str src_fname: name of file as stored in src_dir.

        If the source file is provided locally in the data_path directory,
        it's unnecessary to provide `src_dir` and `src_fname` args.
        """
        self.database = database
        self._data_path = data_path
        self._src_server = src_server
        self._src_dir_path = src_dir_path
        self._src_fname = src_fname
        self._added_ids = []

    def perform_etl(self) -> List[str]:
        """Public-facing method to initiate ETL procedures on given data.

        :return: List of concept IDs which were successfully processed and
            uploaded.
        """
        self._extract_data()
        self._load_meta()
        self._transform_data()
        return self._added_ids

    def _download_data(self, data_path: Path):
        """Download source data from default location."""
        logger.info('Downloading ChemIDplus data...')
        outfile_path = data_path / self._src_fname
        try:
            with FTP(self._src_server) as ftp:
                ftp.login()
                logger.debug('FTP login successful.')
                ftp.cwd(self._src_dir_path)
                with open(outfile_path, 'wb') as fp:
                    ftp.retrbinary(f'RETR {self._src_fname}', fp.write)
            logger.info('Downloaded ChemIDplus source file.')
        except TimeoutError:
            logger.error('Connection to EBI FTP server timed out.')

        parser = ET.iterparse(outfile_path, ('start', 'end'))
        date = next(parser)[1].attrib['date']
        version = date.replace('-', '')
        outfile_path.rename(data_path / f'chemidplus_{version}.xml')
        logger.info('Finished downloading ChemIDplus data')

    def _extract_data(self):
        """Acquire ChemIDplus dataset.

        :arg pathlib.Path data_path: directory containing source data
        """
        self._data_path.mkdir(exist_ok=True, parents=True)
        dir_files = list(self._data_path.iterdir())

        if len(dir_files) == 0:
            file = self._get_file(self._data_path)
        else:
            file = sorted([f for f in dir_files
                           if f.name.startswith('chemidplus')])
            if not file:
                file = self._get_file(self._data_path)

        self._data_src = file[-1]
        self._version = self._data_src.stem.split('_')[1]

    def _get_file(self, data_dir):
        self._download_data(self._data_path)
        dir_files = list(data_dir.iterdir())
        return sorted([f for f in dir_files
                       if f.name.startswith('chemidplus')])

    @staticmethod
    def parse_xml(path: Path, tag: str):
        """Parse XML file and retrieve elements with matching tag value.
        :param Path path: path to XML file
        :param str tag: XML tag
        :return: generator yielding elements of corresponding tag
        """
        context = iter(ET.iterparse(path, events=('start', 'end')))
        _, root = next(context)
        for event, elem in context:
            if event == 'end' and elem.tag == tag:
                yield elem
                root.clear()

    def _transform_data(self):
        """Open dataset and prepare for loading into database."""
        parser = self.parse_xml(self._data_src, 'Chemical')
        with self.database.therapies.batch_writer() as batch:
            for chemical in parser:
                if 'displayName' not in chemical.attrib:
                    continue

                # initial setup and get label
                display_name = chemical.attrib['displayName']
                if not display_name or not re.search(TAGS_REGEX, display_name):
                    continue
                label = re.sub(TAGS_REGEX, '', display_name)
                params = {
                    'label': label
                }

                # get concept ID
                reg_no = chemical.find('NumberList').find("CASRegistryNumber")
                if not reg_no:
                    continue
                params['concept_id'] = f'{NamespacePrefix.CASREGISTRY.value}:{reg_no.text}'  # noqa: E501

                # get aliases
                aliases = []
                label_l = label.lower()
                name_list = chemical.find('NameList')
                if name_list:
                    for name in name_list.findall('NameOfSubstance'):
                        text = name.text
                        if text != display_name and text.lower() != label_l:
                            aliases.append(re.sub(TAGS_REGEX, '', text))
                params['aliases'] = aliases

                # get xrefs and associated_with
                params['xrefs'] = []
                params['associated_with'] = []
                locator_list = chemical.find('LocatorList')
                if locator_list:
                    for loc in locator_list.findall('InternetLocator'):
                        if loc.text == 'DrugBank':
                            db = f'{NamespacePrefix.DRUGBANK.value}:{loc.attrib["url"].split("/")[-1]}'  # noqa: E501
                            params['xrefs'].append(db)
                        elif loc.text == 'FDA SRS':
                            fda = f'{NamespacePrefix.FDA.value}:{loc.attrib["url"].split("/")[-1]}'  # noqa: E501
                            params['associated_with'].append(fda)

                # double-check and load full record
                assert Drug(**params)
                self._load_record(batch, params)

    def _load_record(self, batch: BatchWriter, record: Dict):
        """Load individual record into database.

        :param boto3.dynamodb.table.BatchWriter batch: dynamodb batch writer
            instance
        :param therapy.schemas.Drug record: complete drug record to upload
        """
        concept_id_ref = record['concept_id'].lower()

        if record.get('label'):
            batch.put_item(Item={
                'label_and_type': f'{record["label"].lower()}##label',
                'concept_id': concept_id_ref,
                'src_name': SourceName.CHEMIDPLUS.value,
                'item_type': 'label',
            })
        else:
            del record['label']

        for field_type, field in (('alias', 'aliases'),
                                  ('xref', 'xrefs'),
                                  ('associated_with', 'associated_with')):
            values = record.get(field)
            if values:
                keys = {value.casefold() for value in values}
                if field == 'aliases' and len(keys) > 20:
                    del record['aliases']
                    continue
                for key in keys:
                    pk = f"{key}##{field_type}"
                    batch.put_item(Item={
                        'label_and_type': pk,
                        'concept_id': concept_id_ref,
                        'src_name': SourceName.CHEMIDPLUS.value,
                        'item_type': field_type,
                    })
            else:
                del record[field]

        record['src_name'] = SourceName.CHEMIDPLUS.value
        record['label_and_type'] = f'{concept_id_ref}##identity'
        record['item_type'] = 'identity'
        batch.put_item(Item=record)
        self._added_ids.append(record['concept_id'])

    def _load_meta(self):
        """Add source metadata."""
        meta = SourceMeta(data_license="custom",
                          data_license_url="https://www.nlm.nih.gov/databases/download/terms_and_conditions.html",  # noqa: E501
                          version=self._version,
                          data_url="ftp://ftp.nlm.nih.gov/nlmdata/.chemidlease/",  # noqa: E501
                          rdp_url=None,
                          data_license_attributes=DataLicenseAttributes(
                              non_commercial=False,
                              share_alike=False,
                              attribution=True
                          ))
        item = dict(meta)
        item['src_name'] = SourceName.CHEMIDPLUS.value
        self.database.metadata.put_item(Item=item)



"""Create concept groups and merged records."""
from therapy import PROHIBITED_SOURCES
from therapy.schemas import SourcePriority
from therapy.database import Database
from typing import Set, Dict
import logging
from timeit import default_timer as timer

logger = logging.getLogger('therapy')
logger.setLevel(logging.DEBUG)


class Merge:
    """Handles record merging."""

    def __init__(self, database: Database):
        """Initialize Merge instance.

        :param Database database: db instance to use for record retrieval
            and creation.
        """
        self._database = database
        self._groups = {}  # dict keying concept IDs to group Sets

    def create_merged_concepts(self, record_ids: Set[str]):
        """Create concept groups, generate merged concept records, and
        update database.

        :param Set[str] record_ids: concept identifiers from which groups
            should be generated. Should *not* include any records from
            excluded sources.
        """
        logger.info('Generating record ID sets...')
        start = timer()
        for record_id in record_ids:
            new_group = self._create_record_id_set(record_id)
            if new_group:
                for concept_id in new_group:
                    self._groups[concept_id] = new_group
        end = timer()
        logger.debug(f'Built record ID sets in {end - start} seconds')

        logger.info('Creating merged records and updating database...')
        uploaded_ids = set()
        start = timer()
        for record_id, group in self._groups.items():
            if record_id in uploaded_ids:
                continue
            merged_record = self._generate_merged_record(group)

            # add group merger item to DB
            self._database.add_record(merged_record, 'merger')

            # add updated references
            for concept_id in group:
                if not self._database.get_record_by_id(concept_id, False):
                    logger.error(f"Updating nonexistent record: {concept_id} "
                                 f"for {merged_record['label_and_type']}")
                else:
                    merge_ref = merged_record['concept_id'].lower()
                    self._database.update_record(concept_id, 'merge_ref',
                                                 merge_ref)
            uploaded_ids |= group
        logger.info('Merged concept generation successful.')
        end = timer()
        logger.debug(f'Generated and added concepts in {end - start} seconds')

    def _get_xrefs(self, record: Dict) -> Set[str]:
        """Extract references to entries in other sources from a record.
        Crucially, filter to allowed sources only.

        :param Dict record: record to process
        :return: Set of xref values
        :rtype: Set
        """
        xrefs = set()
        for xref in record.get('xrefs', []):
            if xref.split(':')[0] not in PROHIBITED_SOURCES:
                xrefs.add(xref)
        return xrefs

    def _create_record_id_set(self, record_id: str,
                              observed_id_set: Set = set()) -> Set[str]:
        """Create concept ID group for an individual record ID.

        :param str record_id: concept ID for record to build group from
        :return: set of related identifiers pertaining to a common concept.
        """
        if record_id in self._groups:
            return self._groups[record_id]
        else:
            db_record = self._database.get_record_by_id(record_id)
            if not db_record:
                # attempt RxNorm brand lookup
                brand_lookup = self._database.get_records_by_type(record_id,
                                                                  'rx_brand')
                if len(brand_lookup) == 1:
                    lookup_id = brand_lookup[0]['concept_id']
                    db_record = self._database.get_record_by_id(lookup_id,
                                                                False)
                    if db_record:
                        record_id = db_record['concept_id']
                        return self._create_record_id_set(record_id,
                                                          observed_id_set)
                    else:
                        return observed_id_set - {record_id}

                else:
                    logger.warning(f"Record ID set creator could not resolve "
                                   f"lookup for {record_id} in ID set: "
                                   f"{observed_id_set}")
                    return observed_id_set - {record_id}

            local_id_set = self._get_xrefs(db_record)
            if not local_id_set:
                return observed_id_set | {db_record['concept_id']}
            merged_id_set = {record_id} | observed_id_set
            for local_record_id in local_id_set - observed_id_set:
                merged_id_set |= self._create_record_id_set(local_record_id,
                                                            merged_id_set)
            return merged_id_set

    def _generate_merged_record(self, record_id_set: Set[str]) -> Dict:
        """Generate merged record from provided concept ID group.
        Where attributes are sets, they should be merged, and where they are
        scalars, assign from the highest-priority source where that attribute
        is non-null.

        Priority is:
        RxNorm > NCIt > HemOnc.org > DrugBank > ChemIDplus > Wikidata
        ChEMBL records shouldn't be included in `record_id_set`.

        :param Set record_id_set: group of concept IDs
        :return: completed merged drug object to be stored in DB
        """
        records = []
        for record_id in record_id_set:
            record = self._database.get_record_by_id(record_id)
            if record:
                records.append(record)
            else:
                logger.error(f"Merge record generator could not retrieve "
                             f"record for {record_id} in {record_id_set}")

        def record_order(record):
            """Provide priority values of concepts for sort function."""
            src = record['src_name'].upper()
            if src in SourcePriority.__members__:
                source_rank = SourcePriority[src].value
            else:
                raise Exception(f"Prohibited source: {src} in concept_id "
                                f"{record['concept_id']}")
            return (source_rank, record['concept_id'])
        records.sort(key=record_order)

        # initialize merged record
        merged_attrs = {
            'concept_id': records[0]['concept_id'],
            'label': None,
            'aliases': set(),
            'trade_names': set(),
            'associated_with': set(),
            'approval_status': None,
            'approval_year': set(),
            'fda_indication': [],
        }
        if len(records) > 1:
            merged_attrs['xrefs'] = [r['concept_id'] for r in records[1:]]

        # merge from constituent records
        set_fields = ['aliases', 'trade_names', 'associated_with',
                      'approval_year']
        for record in records:
            for field in set_fields:
                merged_attrs[field] |= set(record.get(field, set()))
            chembl_ids = {i for i in record.get('xrefs', [])
                          if i.startswith('chembl')}
            merged_attrs['associated_with'] |= chembl_ids
            if merged_attrs['label'] is None:
                merged_attrs['label'] = record.get('label')
            if merged_attrs['approval_status'] is None:
                merged_attrs['approval_status'] = record.get('approval_status')
            for ind in record.get('fda_indication', []):
                if ind not in merged_attrs['fda_indication']:
                    merged_attrs['fda_indication'].append(ind)

        # clear unused fields
        for field in set_fields + ['fda_indication']:
            field_value = merged_attrs[field]
            if field_value:
                merged_attrs[field] = list(field_value)
            else:
                del merged_attrs[field]
        for field in ['label', 'approval_status']:
            if merged_attrs[field] is None:
                del merged_attrs[field]

        merged_attrs['label_and_type'] = \
            f'{merged_attrs["concept_id"].lower()}##merger'
        merged_attrs['item_type'] = 'merger'
        return merged_attrs
